/**
 * @brief Este projeto demonstra a programacao de um LCD 2x16
 * @author Wu Shin-Ting
 * @date 13/10/2020
 */

#include "derivative.h" /* include peripheral declarations */

#define GPIO_PIN(x) (1<<x)

typedef enum estado_tag {
	DESABILITA,     /**< falso/apaga/desativa/liga */
	HABILITA,      /**< verdadeiro/acende/ativa/fecha */
} estado_type;

typedef struct rgb_bool_tag {
	estado_type vermelho;    
	estado_type verde;        
	estado_type azul;
} rgb_bool_type;

enum cor_tag {
	VERMELHO,
	VERDE,
	AZUL
};

typedef enum bitmap_tag {
	HEART,
	HAPPY,
	BELL,
	NONE
} bitmap_type;

bitmap_type bitmap;

/**
 * @brief atraso em numero de interacoes <i>
 * @param[in] i numero de iteracoes
 */
void delay (uint32_t i)
{
	__asm("outerLoop:;"
				"sub %0, %0, #1;"
				"mov r2, #11;"
			"innerLoop:;"
				"sub r2, r2, #1;"
				"cmp r2, #0;"
				"bne innerLoop;"
				"cmp %0, #0;"
				"bne outerLoop;"
			: 
			: "r" (i)
			);
}

/**
 * @brief inicializa o componente ledRGB
 */
void GPIO_initLedRGB() {
	// Habilita os clocks dos mdulos
	SIM_SCGC5 |= SIM_SCGC5_PORTE_MASK ;    		// para leds

	// Configure pinos do led RGB com estado apagado
	// Funcao GPIO
	PORTE_PCR21 |= PORT_PCR_MUX(0x1);  // vermelho

	PORTE_PCR22 |= PORT_PCR_MUX(0x1);  // verde

	PORTE_PCR23 |= PORT_PCR_MUX(0x1);  // azul
	// Sentido do sinal: saida
	GPIOE_PDDR |= GPIO_PDDR_PDD(GPIO_PIN(21)) |
			GPIO_PDDR_PDD(GPIO_PIN(22)) |
			GPIO_PDDR_PDD(GPIO_PIN(23));
	// Limpar os bits	
	GPIOE_PCOR = GPIO_PCOR_PTCO(GPIO_PIN(21)) |
			GPIO_PCOR_PTCO(GPIO_PIN(22)) |
			GPIO_PCOR_PTCO(GPIO_PIN(23)) ;	
}

/**
 * @brief seta o estado do componente ledRGB
 * @param[in] estadoR HABILITA/DESABILITA led R
 * @param[in] estadoG HABILITA/DESABILITA led G
 * @param[in] estadoB HABILITA/DESABILITA led B
 */
void GPIO_ledRGB(rgb_bool_type cor) {
	if (cor.vermelho == HABILITA)
		GPIOE_PSOR = GPIO_PSOR_PTSO(GPIO_PIN(21));
	else 
		GPIOE_PCOR = GPIO_PCOR_PTCO(GPIO_PIN(21));

	if (cor.verde == HABILITA)
		GPIOE_PSOR = GPIO_PSOR_PTSO(GPIO_PIN(22));
	else
		GPIOE_PCOR = GPIO_PCOR_PTCO(GPIO_PIN(22));

	if (cor.azul == HABILITA)
		GPIOE_PSOR = GPIO_PSOR_PTSO(GPIO_PIN(23));
	else
		GPIOE_PCOR = GPIO_PCOR_PTCO(GPIO_PIN(23));     
}

/**
 * @brief inicializa o componente Botoeiras 
 */
void GPIO_initSwitches () {
	// Configure pinos das botoeiras
	SIM_SCGC5 |= SIM_SCGC5_PORTA_MASK;				// para botoeiras   					// para leds

	//Funcao GPIO
	PORTA_PCR4 &= ~PORT_PCR_MUX(0x7);			// limpar os bits
	PORTA_PCR4 |= PORT_PCR_MUX(0x1);			// antes de setar 0b001

	PORTA_PCR5 |= PORT_PCR_MUX(0x1);

	PORTA_PCR12 |= PORT_PCR_MUX(0x1);

	//Sentido do sinal: entrada
	GPIOA_PDDR &= ~(GPIO_PDDR_PDD(GPIO_PIN(4)) |
			GPIO_PDDR_PDD(GPIO_PIN(5)) |
			GPIO_PDDR_PDD(GPIO_PIN(12)));    
}

/**
 * @brief Habilita interrupcoes da PORTA no NVIC.
 * @param[in] priority level
 */
void GPIO_enableSwInterrupt(uint8_t priority) {
	/* 
	 * "Non-core interrupt source count" (IRQ = Vector Number (46) - 16, 
	 * Table 3-7 de KL-25 Sub-Family Reference Manual) eh habilitado
	 * ao escrever 1 no bit correpondente do registrador NVIC_ISER 
	 * (B3.4.3 de ARMv6-M Architecture Reference Manual) e a solicitacao de 
	 * interrupcao eh limpa ao escrever 1 no bit correspondente do 
	 * registrador NVIC_ICPR (B3.4.6 de ARMv6-M Architecture Reference Manual) 
	 */
	/* PORTA: Vetor 46, IRQ30, prioridade em NVICIPR7[23:16]
	 * Secao 3.3.2.3 em 
	 * <a href="ftp://ftp.dca.fee.unicamp.br/pub/docs/ea871/ARM/KL25P80M48SF0RM.pdf" target=_manual>ftp://ftp.dca.fee.unicamp.br/pub/docs/ea871/ARM/KL25P80M48SF0RM.pdf</a>
	 */
	NVIC_ISER |= 1 << (46-16);            // Habilita interrupcao PORTA
	NVIC_ICPR |= 1 << (46-16);
	NVIC_IPR7 |= NVIC_IP_PRI_30(priority << 6);    // Nivel de prioridade = 3

	// Comfigurar o trigger
	PORTA_PCR4 |= PORT_PCR_ISF_MASK | 
			PORT_PCR_IRQC(0xA);           // "Interrupt on falling edge"
	PORTA_PCR5 |= PORT_PCR_ISF_MASK | 
			PORT_PCR_IRQC(0xA);           // "Interrupt on falling edge"
	PORTA_PCR12 |= PORT_PCR_ISF_MASK | 
			PORT_PCR_IRQC(0xA);            // "Interrupt on falling edge"
}

/**
 * @brief Rotina de servio para PORTA (Vetor 46, IRQ30, prioridade em NVICIPR7[31:30])
 */
void PORTA_IRQHandler(void) {
	rgb_bool_type cor[3]={{HABILITA,DESABILITA,DESABILITA},
			{DESABILITA,HABILITA,DESABILITA},
			{DESABILITA,DESABILITA,HABILITA}};

	if (PORTA_PCR4 & PORT_PCR_ISF_MASK) {
		// liga led vermelho e apaga outros
		GPIO_ledRGB (cor[VERMELHO]);
		// setar o estado do bitmap
		bitmap = HEART;
		// limpa flag
		PORTA_PCR4 |= PORT_PCR_ISF_MASK; 
		return;
	}
	
	if (PORTA_PCR5 & PORT_PCR_ISF_MASK) {
		// liga led verde e apaga outros
		GPIO_ledRGB (cor[VERDE]);
		// setar o estado do bitmap
		bitmap = HAPPY;
		// limpa flag
		PORTA_PCR5 |= PORT_PCR_ISF_MASK;
		return;
	} 
	
	if (PORTA_PCR12 & PORT_PCR_ISF_MASK) {
		// liga led azul e apaga outros
		GPIO_ledRGB (cor[AZUL]);
		// setar o estado do bitmap
		bitmap = BELL;
		// limpa flag
		PORTA_PCR12 |= PORT_PCR_ISF_MASK;
		
		return;
	}
}

/**
 * @brief Funcao do byte
 */
typedef enum lcd_RS_tag {
	COMANDO,  /**< Comando */
	DADO	  /**< Dado */
} lcd_RS_type;

/**
 * @brief Inicializa os pinos conectados ao latch 74573 onde estao conectados os 8 leds vermelhos
 */
void GPIO_initLCD(){
	SIM_SCGC5 |= SIM_SCGC5_PORTC_MASK ;     ///< Habilita o clock da PORTA C

	// Configura os pinos PORTC[10:0] como GPIO de saida 
	PORTC_PCR0 |= PORT_PCR_MUX(0x1);       ///< D0-D7 dos dados (GPIO)
	PORTC_PCR1 |= PORT_PCR_MUX(0x1);
	PORTC_PCR2 |= PORT_PCR_MUX(0x1);
	PORTC_PCR3 |= PORT_PCR_MUX(0x1);
	PORTC_PCR4 |= PORT_PCR_MUX(0x1);
	PORTC_PCR5 |= PORT_PCR_MUX(0x1);
	PORTC_PCR6 |= PORT_PCR_MUX(0x1);
	PORTC_PCR7 |= PORT_PCR_MUX(0x1);
	
	PORTC_PCR8 |= PORT_PCR_MUX(0x1);       ///< RS do LCD
	PORTC_PCR9 |= PORT_PCR_MUX(0x1);       ///< E do LCD

	GPIOC_PDDR |= GPIO_PDDR_PDD(GPIO_PIN(9)|GPIO_PIN(8) |GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0));       ///< Configura como saidas
	GPIOC_PDOR &= ~GPIO_PDOR_PDO(GPIO_PIN(9) |GPIO_PIN(8)|GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0)); ///< Inicializa os pinos com 0
}

/**
 * @brief Escrever nos pinos de dados um caractere de 8 bits
 * @param[in] c caractere
 */
void GPIO_setByte(char c) {
	// colocar os dados nos pinos de dados DO-D7
	GPIOC_PDOR &= 0xFFFFFF00; 		// zerar o byte menos significativo
	GPIOC_PDOR |= (unsigned int) c; // setas os bits em 1 no byte menos significativo	
}

/**
 * @brief Gerar o pulso de Enable de ~450ns
 */
void GPIO_enableE()
{
	// gerar um pulso  (~ 450 ns)
	GPIOC_PSOR = GPIO_PIN (9);		// setar em 1
	delay(1);
	GPIOC_PCOR = GPIO_PIN (9);		// resetar em 0
}

/**
 * @brief Setar o tipo de transferencia
 * @param[in] i tipo
 */
void GPIO_setRS (lcd_RS_type i) {
	if(i == COMANDO) 
		GPIOC_PCOR = GPIO_PIN(8);     ///< Seta o LCD no modo de comando		
	else if (i == DADO) 
		GPIOC_PSOR = GPIO_PIN(8);    ///< Seta o LCD no modo de dados
}

/**
 * @brief Transferir um byte ao LCD
 * @param[in] c byte
 * @param[in] t tempo de processamento em multiplos de 2 us
 */
void escreveLCD(char c, uint32_t t)
{
	GPIO_setByte(c);
	GPIO_enableE();
	delay(t);
}

/**
 * @brief Inicializar o LCD
 */
void initLCD () {
	//espera por mais de 30ms = 30000us/2 = 15000
	delay(15000);
	
	GPIO_setRS(COMANDO);
	escreveLCD(0x38, 20);			//Function Set: 39us/2 -> 20
	escreveLCD(0x0C, 20);		    //Display ON/OFF Control: 39us/2 -> 20
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
	escreveLCD(0x06, 20);			//Entry mode set: 39us/2 -> 20
}

/**
 * @brief Limpe o display
 */
void limpeDisplayLCD () {
	GPIO_setRS(COMANDO);
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
}

/**
 * @brief Escrever uma mensagem str no LCD a partir do endereco especificado
 * @param[in] end endereco DDRAM
 * @param[in] str mensagem
 */
void escreveMensagem (uint8_t end, char* str)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b10000000 | end; // seta o bit 7 em 1
	GPIO_setRS(COMANDO);			// comando
	escreveLCD(tmp, 20);			// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	while (*str) {
		escreveLCD(*str, 20);
		str++;
	}
}

/**
 * @brief Escrever um bitmap na memoria CGRAM a partir do endereco especificado
 * @param[in] end endereco CGRAM
 * @param[in] str bitmap
 */
void escreveBitmap (uint8_t end, char* bitmap)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b01000000 | (end*8); // seta o bit 7 em 0 e bit 6 em 1
	GPIO_setRS(COMANDO);				// comando
	escreveLCD(tmp, 20);				// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	uint8_t i;
	for (i=0; i< 8; i++) {
		escreveLCD(bitmap[i], 20);
	}
}

int main(void)
{
	char happy[8]={0x00,0x0A,0x0A,0x00,0x11,0x0F,0x00,0x00};

	char heart[8]={0x00,0x0A,0x15,0x11,0x0A,0x04,0x00,0x00};

	char bell[8] = {0x00,0x04,0x0E,0x0E,0x0E,0x1F,0x00,0x04};
	
	GPIO_initLCD();
	
	initLCD();
	
	escreveBitmap(HEART,heart);
	escreveBitmap(HAPPY,happy);
	escreveBitmap(BELL,bell);
	
	GPIO_initLedRGB();
	GPIO_initSwitches();
	GPIO_enableSwInterrupt(1);
	
	uint8_t tmp;

	bitmap = NONE;
	
	for(;;) {
		switch (bitmap) {
		case HEART:
			limpeDisplayLCD();
			//Posicione o cursor no canto superior direito
			tmp = 0b10000000 | 0x0F; // seta o bit 7 em 1
			GPIO_setRS(COMANDO);			// comando
			escreveLCD(tmp, 20);			// seta o endereco do cursor

			//Chaveia para modo DADO
			GPIO_setRS (DADO);
		   	escreveLCD(HEART, 20);
		   	
		   	//Escreve a mensagem
		   	escreveMensagem (0x40, "VERMELHO");
		   	
		   	bitmap = NONE;
			break;
		case HAPPY:
			limpeDisplayLCD();
			//Posicione o cursor no canto superior direito
			tmp = 0b10000000 | 0x0F; // seta o bit 7 em 1
			GPIO_setRS(COMANDO);			// comando
			escreveLCD(tmp, 20);			// seta o endereco do cursor

			//Chaveia para modo DADO
			GPIO_setRS (DADO);
		   	escreveLCD(HAPPY, 20);
		   	
		   	//Escreve a mensagem
		   	escreveMensagem (0x40, "VERDE");

		   	bitmap = NONE;
			break;
		case BELL:
			limpeDisplayLCD();
			//Posicione o cursor no canto superior direito
			tmp = 0b10000000 | 0x0F; // seta o bit 7 em 1
			GPIO_setRS(COMANDO);			// comando
			escreveLCD(tmp, 20);			// seta o endereco do cursor

			//Chaveia para modo DADO
			GPIO_setRS (DADO);
		   	escreveLCD(BELL, 20);
		   	
		   	//Escreve a mensagem
		   	escreveMensagem (0x40, "AZUL");
		   	
		   	bitmap = NONE;
			break;
		case NONE:
			break;
		}
	}
	
	return 0;
}
